import logging

from .servicex_databinder import DataBinder # NOQA

__version__ = '0.5.0'

logging.basicConfig(format="%(levelname)s - %(message)s")
logging.getLogger(__name__).setLevel(logging.INFO)



import yaml
from pathlib import Path
from typing import Any, Dict
from shutil import rmtree, copy

import pyarrow.parquet as pq
import awkward as ak
import uproot

import logging
log = logging.getLogger(__name__)


class OutputHandler():

    def __init__(self, config: Dict[str, Any]) -> None:
        self._config = config
        self._outputformat \
            = self._config.get('General')['OutputFormat'].lower()
        """
        Prepare output path dictionary
        """
        out_paths = {}
        samples = []
        [samples.append(sample['Name'])
            for sample in config['Sample'] if sample['Name'] not in samples]
        for sample in samples:
            out_paths[sample] = {}
        for sample in config['Sample']:
            # if sample['Transformer'] == "uproot":
            if 'Tree' in sample.keys():
                for tree in sample['Tree'].split(','):
                    out_paths[sample['Name']][tree.strip()] = []

        self.out_paths_dict = out_paths

        """
        Create output directory
        """
        if 'OutputDirectory' in self._config['General'].keys():
            self.output_path = Path(
                self._config['General']['OutputDirectory']
                ).absolute()
            self.output_path.mkdir(parents=True, exist_ok=True)
        else:
            self.output_path = Path('ServiceXData').absolute()
            self.output_path.mkdir(parents=True, exist_ok=True)

    def copy_to_target(self, delivery_setting, req, files):
        if req['codegen'] == "uproot":
            target_path = Path(self.output_path, req['Sample'], req['tree'])
            delivery_info = (f"  {req['Sample']} | "
                             f"{req['tree']} | "
                             f"{str(req['dataset'])[:100]}")
        elif req['codegen'] == "atlasr21" or req['codegen'] == "python":
            target_path = Path(self.output_path, req['Sample'])
            delivery_info = (f"  {req['Sample']} | "
                             f"{str(req['dataset'])[:100]}")

        if delivery_setting == 1 or delivery_setting == 2:
            if target_path.exists():
                servicex_files = {Path(file).name for file in files}
                local_files = {
                    Path(file).name for file in list(target_path.glob("*"))
                    }
                servicex_data_path = Path(files[0]).parents[0]
                # one RucioDID for this sample and files are already there
                if servicex_files == local_files:
                    log.info(f"{delivery_info} is already delivered")
                else:
                    # copy files in servicex but not in local
                    files_not_in_local = servicex_files.difference(local_files)
                    if files_not_in_local:
                        for file in files_not_in_local:
                            copy(Path(servicex_data_path, file),
                                 Path(target_path, file))
                        log.info(f"{delivery_info} is delivered")
                    else:
                        log.info(f"{delivery_info} is already delivered")
            else:
                target_path.mkdir(parents=True, exist_ok=True)
                for file in files:
                    outfile = Path(target_path, Path(file).name)
                    copy(file, outfile)
                log.info(f"{delivery_info} is delivered")
        elif delivery_setting == 3 or delivery_setting == 4:
            log.info(f"{delivery_info} is cached locally")
        elif delivery_setting == 5 or delivery_setting == 6:
            log.info(f"{delivery_info} is available at the object store")

    def parquet_to_root(self, tree_name, pq_file, root_file):
        """
        Write ROOT ntuple from parquet file
        """
        if pq.read_metadata(pq_file).num_rows == 0:
            pass
        else:
            outfile = uproot.recreate(root_file)
            tree_dict = {}
            ak_arr = ak.from_parquet(pq_file)
            for field in ak_arr.fields:
                tree_dict[field] = ak_arr[field]
            outfile[tree_name] = tree_dict
            outfile.close()

    def update_output_paths_dict(
            self,
            req,
            files,
            delivery_setting: int
            ):
        """
        Update dictionary of outfile paths
        """
        if req['codegen'] == "uproot":
            target_path = Path(self.output_path, req['Sample'], req['tree'])
            paths_in_output_dict = \
                self.out_paths_dict[req['Sample']][req['tree']]
        elif req['codegen'] == "atlasr21" or req['codegen'] == "python":
            target_path = Path(self.output_path, req['Sample'])
            paths_in_output_dict = self.out_paths_dict[req['Sample']]

        # Update file path if deliver to localpath
        if delivery_setting == 1 or delivery_setting == 2:
            new_files = [
                str(Path(target_path, Path(file).name))
                for file in files
                ]
        elif delivery_setting == 5 or delivery_setting == 6:
            new_files = [file._url for file in files]
        else:
            new_files = [str(file) for file in files]

        # Update output_dict
        if paths_in_output_dict:
            output_dict = list(set(paths_in_output_dict + new_files))
        else:
            output_dict = new_files

        if req['codegen'] == "uproot":
            self.out_paths_dict[req['Sample']][req['tree']] = output_dict
        elif req['codegen'] == "atlasr21" or req['codegen'] == "python":
            self.out_paths_dict[req['Sample']] = output_dict

    def add_local_output_paths_dict(self):
        local_samples = [sample for sample in self._config.get('Sample')
                         if 'LocalPath' in sample.keys()]
        for sample in local_samples:
            if 'Tree' in sample.keys():
                for tree, fpath in zip(sample['Tree'].split(','),
                                       sample['LocalPath'].split(',')):
                    tree = tree.strip()
                    fpath = fpath.strip()
                    self.out_paths_dict[sample['Name']][tree] \
                        = [str(Path(f)) for f in Path(fpath).glob("*")]
                    log.info(f"  {sample['Name']} "
                             f"| {tree} | {fpath} is from local path")
            else:
                for fpath in sample['LocalPath'].split(','):
                    fpath = fpath.strip()
                    self.out_paths_dict[sample['Name']] = \
                        [str(Path(f)) for f in Path(fpath).glob("*")]
                    log.info(f"  {sample['Name']} "
                             f"| {fpath} is from local path")

    def write_output_paths_dict(self, out_paths_dict):
        """
        Write yaml of output paths
        """
        if 'WriteOutputDict' in self._config['General'].keys():
            file_out_paths = \
                (f"{self.output_path}/"
                 f"{self._config['General']['WriteOutputDict']}.yml")
            with open(file_out_paths, 'w') as f:
                log.debug("YAML file containing delivered file paths: "
                          f"{f.name}")
                yaml.dump(out_paths_dict, f, default_flow_style=False)
            log.info("YAML file containing delivered file paths: "
                     f"{file_out_paths}")
        else:
            for yl in list(Path(self.output_path).glob("*yml")):
                Path.unlink(yl)

    def clean_up_files_not_in_requests(self, out_paths_dict):

        samples_in_requests = list(out_paths_dict.keys())
        samples_local = [sa.name for sa in self.output_path.iterdir()
                         if sa.is_dir()]
        for sample in samples_local:
            if not (sample in samples_in_requests):
                rmtree(Path(self.output_path, sample))
            else:
                if list(Path(self.output_path, sample).iterdir())[0].is_dir():
                    for tree in out_paths_dict[sample].keys():
                        for tree in [tr.name for tr in
                                     Path(self.output_path, sample).iterdir()
                                     if tr.is_dir()]:
                            if tree in out_paths_dict[sample].keys():
                                files_local = set(Path(self.output_path,
                                                       sample, tree).glob("*"))
                                files_request = set(
                                    [Path(item)
                                     for item in out_paths_dict[sample][tree]]
                                    )
                                for tbd in \
                                        files_local.difference(files_request):
                                    Path.unlink(tbd)
                            else:
                                rmtree(Path(self.output_path, sample, tree))
                else:
                    files_local = set(Path(self.output_path, sample).glob("*"))
                    files_request = set(
                        [Path(item) for item in out_paths_dict[sample]]
                        )
                    for tbd in files_local.difference(files_request):
                        Path.unlink(tbd)

        return



from typing import Any, Dict, List
import tcut_to_qastle as tq
import qastle
import ast
import logging
from base64 import b64encode
from func_adl_servicex import ServiceXSourceXAOD # NOQA
from servicex import ServiceXDataset # NOQA

log = logging.getLogger(__name__)


class ServiceXRequest():
    """
    Prepare ServiceX requests
    """
    def __init__(self, config: Dict[str, Any]) -> None:
        self._config = config

    def get_requests(self) -> List:
        log.debug(f"ServiceX backend: "
                  f"{self._config.get('General')['ServiceXName']}")
        list_requests = []
        for sample in self._config.get('Sample'):
            list_requests.append(self._build_request(sample))
        # flatten nested lists
        flist_requests = [request for x in list_requests for request in x]
        log.debug("number of total ServiceX requests in the config: "
                  f"{len(flist_requests)}")
        log.debug(f"ServiceX requests in the config: {flist_requests}")
        return flist_requests

    def _build_request(self, sample: Dict) -> Dict:
        """
        Return a list containing ServiceX request(s) of the given sample
        """
        requests_sample = []

        if 'RucioDID' in sample.keys():
            dids = sample['RucioDID'].split(',')
            if sample['Transformer'] == "uproot":
                trees = sample['Tree'].split(',')
                log.debug(f"  Sample {sample['Name']} has "
                          f"{len(dids)} DID(s) and {len(trees)} Tree(s)")
            else:
                trees = ['dummy']
                log.debug(f"  Sample {sample['Name']} has {len(dids)} DID(s)")

            for tree in trees:
                for did in dids:
                    requests_sample.append(
                        {
                            'Sample': sample['Name'],
                            'dataset': did.strip(),
                            'type': sample['Type'],
                            'codegen': sample['Transformer'],
                            'tree': tree.strip(),
                            'query': self._build_query(sample, tree.strip())
                        }
                    )
        elif 'XRootDFiles' in sample.keys():
            xrootd_filelist = [file.strip()
                               for file in sample['XRootDFiles'].split(",")]
            if sample['Transformer'] == "uproot":
                trees = sample['Tree'].split(',')
                log.debug(f"  Sample {sample['Name']} has "
                          f"{len(xrootd_filelist)} file(s) and "
                          f"{len(trees)} Tree(s)")
            else:
                trees = ['dummy']
                log.debug(f"  Sample {sample['Name']} has "
                          f"{len(xrootd_filelist)} file(s)")
            for tree in trees:
                requests_sample.append(
                    {
                        'Sample': sample['Name'],
                        'dataset': xrootd_filelist,
                        'type': sample['Type'],
                        'codegen': sample['Transformer'],
                        'tree': tree.strip(),
                        'query': self._build_query(sample, tree)
                    }
                )
        return requests_sample

    def _build_query(self, sample: Dict, tree: str) -> str:
        """
        Get query for each sample
        Option Columns for TCut syntax
        Option FuncADL for func-adl syntax
        """
        if sample['Transformer'] == "uproot":
            if 'Columns' in sample:
                if ('Filter' not in sample) or (sample['Filter'] is None):
                    sample['Filter'] = ''
                # else:
                try:
                    query = tq.translate(
                        tree,
                        sample['Columns'],
                        sample['Filter']
                    )
                    return query
                except Exception:
                    log.exception(
                        "Exception occured for the query "
                        f"of Sample {sample['Name']}"
                        )
            elif 'FuncADL' in sample:
                query = ("EventDataset('ServiceXDatasetSource', "
                         f"'{tree}')." + sample['FuncADL'])
                try:
                    qastle_query = qastle.python_ast_to_text_ast(
                        qastle.insert_linq_nodes(ast.parse(query)))
                    return qastle_query
                except Exception:
                    log.exception(
                        "Exception occured for the query "
                        f"of Sample {sample['Name']}"
                        )
        elif sample['Transformer'] == "atlasr21":
            query = "ServiceXSourceXAOD(ServiceXDataset('', backend_name='" \
                + self._config.get('General')['ServiceXName'] + "'))." \
                + sample['FuncADL']
            try:
                o = eval(query)
                qastle_query = qastle.python_ast_to_text_ast(o._q_ast)
                return qastle_query
            except Exception:
                log.exception(
                    "Exception occured for the query "
                    f"of Sample {sample['Name']}"
                    )
        elif sample['Transformer'] == "python":
            query = sample['Function']
            return b64encode(query.encode("utf-8")).decode("utf-8")



from typing import Union, Dict, Any
from pathlib import Path
import asyncio
from threading import Thread

from .configuration import LoadConfig
from .request import ServiceXRequest
from .get_servicex_data import DataBinderDataset
from .output_handler import OutputHandler

import logging
log = logging.getLogger(__name__)


class DataBinder:
    """
    Manage and categorize numerous ServiceX data from a configuration file
    """

    def __init__(self, config: Union[str, Path, Dict[str, Any]]):
        self._config = LoadConfig(config)
        self._requests = ServiceXRequest(self._config).get_requests()
        self._sx_db = DataBinderDataset(self._config, self._requests)

        log.info(f"  {len(self._config.get('Sample'))} Samples"
                 f" and {len(self._requests)} ServiceX requests")

    def deliver(self, overall_progress_only: bool = False) -> Dict:

        out_paths_dict = asyncio.run(
                self._sx_db.get_data(overall_progress_only)
            )

        x = Thread(target=OutputHandler(self._config)
                   .clean_up_files_not_in_requests, args=(out_paths_dict,))
        x.start()

        if len(self._sx_db.failed_request):
            log.warning(f"{len(self._sx_db.failed_request)} "
                        "failed delivery request(s)")
            log.warning("get_failed_requests() for detail of failed requests")

        return out_paths_dict

    def get_failed_requests(self):
        return self._sx_db.failed_request



from unittest import mock
import pytest

from servicex_databinder import configuration

@mock.patch("servicex_databinder.configuration._validate_config")
def test_load_config(mock_validation):
    conf = configuration._load_config("config_example_uproot.yml")
    assert isinstance(conf, dict)
    mock_validation.assert_called_once()

    with pytest.raises(FileNotFoundError):
        conf = configuration._load_config("none.yml")

def test_validate_config():
    config_without_general = {}
    with pytest.raises(KeyError):
        configuration._validate_config(config_without_general)

    config_without_servicexbackendname = {"General": {}, 
                            "Sample": [{},]}
    with pytest.raises(KeyError):
        configuration._validate_config(config_without_servicexbackendname)

    config_without_wrong_backendname = {
        "General": {
            "ServiceXBackendName": "wrong_backend_name",
        },
        "Sample": [{},]}
    with pytest.raises(ValueError):
        configuration._validate_config(config_without_wrong_backendname)
    
    config_without_outputdirectory = {
        "General": {
            "ServiceXBackendName": "uproot",
        },
        "Sample": [{},]}
    with pytest.raises(KeyError):
        configuration._validate_config(config_without_outputdirectory)

    config_without_outputformat = {
        "General": {
            "ServiceXBackendName": "uproot",
            "OutputDirectory": "a",
        },
        "Sample": [{},]}
    with pytest.raises(KeyError):
        configuration._validate_config(config_without_outputformat)

    config_wrong_outputformat = {
        "General": {
            "ServiceXBackendName": "uproot",
            "OutputDirectory": "a",
            "OutputFormat": "pandas",
        },
        "Sample": [{},]}
    with pytest.raises(ValueError):
        configuration._validate_config(config_wrong_outputformat)

    config_without_rucio_did = {
        "General": {
            "ServiceXBackendName": "uproot",
            "OutputDirectory": "a",
            "OutputFormat": "parquet",
        },
        "Sample": [{
            "Name": "ttH",
        }]
    }
    with pytest.raises(KeyError):
        configuration._validate_config(config_without_rucio_did)

    config_without_rucio_scope = {
        "General": {
            "ServiceXBackendName": "uproot",
            "OutputDirectory": "a",
            "OutputFormat": "parquet",
        },
        "Sample": [{
            "Name": "ttH",
            "RucioDID": "user.kchoi:user.kchoi.A, user.kchoi.B",
        }]
    }
    with pytest.raises(ValueError):
        configuration._validate_config(config_without_rucio_scope)

    config_tree_not_with_uproot = {
        "General": {
            "ServiceXBackendName": "xaod",
            "OutputDirectory": "a",
            "OutputFormat": "root",
        },
        "Sample": [{
            "Name": "ttH",
            "RucioDID": "user.kchoi:user.kchoi.A, user.kchoi:user.kchoi.B",
            "Tree": "nominal",
        }]
    }
    with pytest.raises(KeyError):
        configuration._validate_config(config_tree_not_with_uproot)

    config_columns_with_funcadl = {
        "General": {
            "ServiceXBackendName": "xaod",
            "OutputDirectory": "a",
            "OutputFormat": "root",
        },
        "Sample": [{
            "Name": "ttH",
            "RucioDID": "user.kchoi:user.kchoi.A, user.kchoi:user.kchoi.B",
            "Columns": "jet_pt",
            "FuncADL": "Select()",
        }]
    }
    with pytest.raises(KeyError):
        configuration._validate_config(config_columns_with_funcadl)

    config_filter_with_funcadl = {
        "General": {
            "ServiceXBackendName": "xaod",
            "OutputDirectory": "a",
            "OutputFormat": "root",
        },
        "Sample": [{
            "Name": "ttH",
            "RucioDID": "user.kchoi:user.kchoi.A, user.kchoi:user.kchoi.B",
            "Filter": "jet_pt>15e3",
            "FuncADL": "Select()",
        }]
    }
    with pytest.raises(KeyError):
        configuration._validate_config(config_filter_with_funcadl) 

    config_valid_uproot ={
        "General": {
            "ServiceXBackendName": "uproot",
            "OutputDirectory": "a",
            "OutputFormat": "parquet",
        },
        "Sample": [{
            "Name": "ttH",
            "RucioDID": "user.kchoi:user.kchoi.A, user.kchoi:user.kchoi.B",
            "FuncADL": "Select()",
        }]
    }
    assert configuration._validate_config(config_valid_uproot)




# def test_validate_config_for_uproot():
#     config_valid_uproot ={
#         "General": {
#             "ServiceXBackendName": "uproot_test",
#             "OutputDirectory": "a",
#             "OutputFormat": "parquet",
#         },
#         "Sample": [{
#             "Name": "ttH",
#             "RucioDID": "user.kchoi:user.kchoi",
#             "FuncADL": "Select()",
#         }]
#     }
#     assert configuration._validate_config(config_valid_uproot)

# def test_validate_config_for_xaod():
#     config_valid_xaod ={
#         "General": {
#             "ServiceXBackendName": "xaod_test",
#             "OutputDirectory": "a",
#             "OutputFormat": "root",
#         },
#         "Sample": [{
#             "Name": "ttH",
#             "RucioDID": "user.kchoi:user.kchoi",
#             "FuncADL": 1,
#         }]
#     }
#     assert configuration._validate_config(config_valid_xaod)


from typing import Any, Dict, List
import logging

from aiohttp import ClientSession
import asyncio
from tqdm.asyncio import tqdm

from servicex import ServiceXDataset, utils, servicex_config
from .output_handler import OutputHandler

import nest_asyncio
nest_asyncio.apply()

log = logging.getLogger(__name__)


class DataBinderDataset:

    def __init__(self, config: Dict[str, Any], servicex_requests: List):
        self._config = config
        self._servicex_requests = servicex_requests
        self._outputformat \
            = self._config.get('General')['OutputFormat'].lower()
        self.transformerImage = None
        self.output_handler = OutputHandler(config)
        self.ignoreCache = False
        if 'IgnoreServiceXCache' in self._config['General'].keys():
            self.ignoreCache = self._config['General']['IgnoreServiceXCache']
        self.failed_request = []
        self.endpoint = servicex_config.ServiceXConfigAdaptor()\
            .get_backend_info(
                self._config['General']['ServiceXName'],
                "endpoint"
                )

    async def deliver_and_copy(self, req, delivery_setting):
        if req['codegen'] == "uproot":
            title = f"{req['Sample']} - {req['tree']}"
        elif req['codegen'] == "atlasr21" or req['codegen'] == "python":
            title = f"{req['Sample']}"

        if self._progresbar:
            callback_factory = None
        else:
            callback_factory = utils._run_default_wrapper

        try:
            async with ClientSession(timeout=3600) as session:
                sx_ds = ServiceXDataset(
                    dataset=req['dataset'],
                    backend_name=self._config['General']['ServiceXName'],
                    backend_type=req['type'],
                    codegen=req['codegen'],
                    # image=self.transformerImage,
                    status_callback_factory=callback_factory,
                    session_generator=session,
                    ignore_cache=self.ignoreCache
                    )
                query = req['query']
                if delivery_setting == 1 or delivery_setting == 3:
                    files = await sx_ds.get_data_parquet_async(
                        query,
                        title=title
                        )
                elif delivery_setting == 2 or delivery_setting == 4:
                    files = await sx_ds.get_data_rootfiles_async(
                        query,
                        title=title
                        )
                elif delivery_setting == 5:
                    files = await sx_ds.get_data_parquet_uri_async(
                        query,
                        title=title
                        )
                elif delivery_setting == 6:
                    files = await sx_ds.get_data_rootfiles_uri_async(
                        query,
                        title=title
                        )

            # Update Outfile paths dictionary
            self.output_handler.update_output_paths_dict(
                req, files, delivery_setting
                )

            self.output_handler.copy_to_target(delivery_setting, req, files)
        except Exception as e:
            self.failed_request.append({"request": req, "error": repr(e)})
            if req['codegen'] == "uproot":
                return ("  Fail to deliver "
                        f"{req['Sample']} | "
                        f"{req['tree']} | "
                        f"{str(req['dataset'])[:100]}")
            elif req['codegen'] == "atlasr21":
                return ("  Fail to deliver "
                        f"{req['Sample']} | "
                        f"{str(req['dataset'])[:100]}")

    async def get_data(self, overall_progress_only):
        log.info(f"Deliver via ServiceX endpoint: {self.endpoint}")

        if self._outputformat == "parquet" and \
                self._config['General']['Delivery'] == "localpath":
            delivery_setting = 1
        elif self._outputformat == "root" and \
                self._config['General']['Delivery'] == "localpath":
            delivery_setting = 2
        elif self._outputformat == "parquet" and \
                self._config['General']['Delivery'] == "localcache":
            delivery_setting = 3
        elif self._outputformat == "root" and \
                self._config['General']['Delivery'] == "localcache":
            delivery_setting = 4
        elif self._outputformat == "parquet" and \
                self._config['General']['Delivery'] == "objectstore":
            delivery_setting = 5
        elif self._outputformat == "root" and \
                self._config['General']['Delivery'] == "objectstore":
            delivery_setting = 6

        self._progresbar = overall_progress_only

        tasks = []

        for req in self._servicex_requests:
            tasks.append(self.deliver_and_copy(req, delivery_setting))

        if overall_progress_only:
            barformat = "{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}]"
            pbar = tqdm(total=len(tasks),
                        unit="request",
                        dynamic_ncols=True,
                        colour='#ffa500',
                        bar_format=barformat,
                        )
        for f in asyncio.as_completed(tasks):
            value = await f
            if overall_progress_only:
                pbar.set_description(value)
                pbar.update()
            else:
                pass

        if overall_progress_only:
            pbar.close()

        self.output_handler.add_local_output_paths_dict()

        if delivery_setting == 1 or delivery_setting == 2:
            log.info(f"Delivered at {self.output_handler.output_path}")

        self.output_handler.write_output_paths_dict(
            self.output_handler.out_paths_dict
            )

        return self.output_handler.out_paths_dict



from servicex import servicex_config
import yaml
import pathlib
from typing import Any, Dict, Union
import logging
log = logging.getLogger(__name__)


def LoadConfig(input_config:
               Union[str, pathlib.Path, Dict[str, Any]]
               ) -> Dict[str, Any]:
    """Loads, validates, and returns a config file from the provided path.
    Args:
        input_config (Union[str, pathlib.Path, Dict[str, Any]]):
            path to config file or config as dict
    Returns:
        Dict[str, Any]: configuration
    """

    if isinstance(input_config, dict):
        _decorate_defaults(input_config)
        _replace_definition_in_sample_block(input_config)
        _validate_config(input_config)
        return _update_backend_per_sample(input_config)
    else:
        file_path = pathlib.Path(input_config)
        log.info(f"Loading DataBinder config file: {file_path}")
        try:
            config = yaml.safe_load(file_path.read_text())
            _decorate_defaults(config)
            _replace_definition_in_sample_block(config)
            _validate_config(config)
            return _update_backend_per_sample(config)
        except Exception:
            raise SyntaxError(
                f"Exception occured while reading config file: {file_path}"
                )


def _replace_definition_in_sample_block(config: Dict[str, Any]):
    flag = False
    if config.get('Definition'):
        for n, sample in enumerate(config.get('Sample')):
            for field, value in sample.items():
                if 'DEF_' in value:
                    for repre, new_str in config.get('Definition').items():
                        if repre in value:
                            log.debug(
                                f"Replace Definition for {sample['Name']} "
                                f"- {field}: {repre} with {new_str}")
                            config.get('Sample')[n][field] \
                                = config.get('Sample')[n][field] \
                                        .replace(repre, new_str)
                            flag = True
        return flag
    else:
        return flag


def _update_backend_per_sample(config: Dict[str, Any]) -> Dict:
    """ from servicex.yaml file """
    backend_type = servicex_config.ServiceXConfigAdaptor()\
        .get_backend_info(config['General']['ServiceXName'], "type")
    if backend_type == "xaod":
        pair = ("xaod", "atlasr21")
    elif backend_type == "uproot":
        pair = ("uproot", "uproot")

    """ from General block """
    if 'Transformer' in config['General'].keys():
        if config['General']['Transformer'] == "atlasr21":
            pair = ("xaod", "atlasr21")
        elif config['General']['Transformer'] == "uproot":
            pair = ("uproot", "uproot")
        elif config['General']['Transformer'] == "python":
            pair = ("uproot", "python")

    """ from Sample block """
    for (idx, sample) in zip(range(len(config['Sample'])), config['Sample']):
        if 'Transformer' in sample.keys():
            if sample['Transformer'] == "atlasr21":
                config['Sample'][idx]['Type'] = "xaod"
            elif sample['Transformer'] == "uproot":
                config['Sample'][idx]['Type'] = "uproot"
            elif sample['Transformer'] == "python":
                config['Sample'][idx]['Type'] = "uproot"
        else:
            config['Sample'][idx]['Type'] = pair[0]
            config['Sample'][idx]['Transformer'] = pair[1]

    return config


def _decorate_defaults(config: Dict[str, Any]) -> Dict:
    if 'Delivery' in config['General'].keys():
        config['General']['Delivery'] = config['General']['Delivery'].lower()
    else:
        config['General']['Delivery'] = 'localpath'
    return config


def _validate_config(config: Dict[str, Any]) -> bool:
    """Returns True if the config file is validated,
    otherwise raises exceptions.
    Checks that the config satisfies the json schema,
    and performs additional checks to
    validate the config further.
    Args:
        config (Dict[str, Any]): configuration
    Raises:
        NotImplementedError
        ValueError
        KeyError
    Returns:
        bool: whether the validation was successful
    """

    # Check Option names
    available_keys = [
        'General', 'ServiceXName', 'OutputDirectory', 'Transformer',
        'OutputFormat', 'WriteOutputDict', 'Name',
        'IgnoreLocalCache', 'Sample', 'RucioDID', 'XRootDFiles', 'Tree',
        'Filter', 'Columns', 'FuncADL', 'LocalPath', 'Definition',
        'ServiceXBackendName', 'IgnoreServiceXCache',
        'Delivery', 'Function'
        ]

    if 'General' not in config.keys() and 'Sample' not in config.keys():
        raise KeyError("You should have 'General' block and "
                       "at least one 'Sample' block in the config")

    keys_in_config = set()

    for item in config['General']:
        keys_in_config.add(item)
    for sample in config['Sample']:
        for item in sample.keys():
            keys_in_config.add(item)
    for key in keys_in_config:
        if key not in available_keys:
            raise KeyError(f"Unknown Option {key} in the config")

    # Check General block option values
    if 'Delivery' in config['General'].keys():
        if config['General']['Delivery'] not in [
                'localpath', 'localcache', 'objectstore']:
            raise ValueError(
                f"Unsupported delivery option: {config['General']['Delivery']}"
                f" - supported options: LocalPath, LocalCache, ObjectStore"
                )

    if ('ServiceXName' not in config['General'].keys()) and \
            ('ServiceXBackendName' not in config['General'].keys()):
        raise KeyError("Option 'ServiceXName' is required in General block")

    if 'OutputFormat' not in config['General'].keys():
        raise KeyError("OutputFormat is required")
    elif config['General']['OutputFormat'].lower() != 'parquet' and \
            config['General']['OutputFormat'].lower() != 'root':
        raise ValueError("OutputFormat can be either parquet or root")

    for sample in config['Sample']:
        if ('RucioDID' not in sample.keys()) \
                and ('XRootDFiles' not in sample.keys()) \
                and ('LocalPath' not in sample.keys()):
            raise KeyError(
                "Please specify a valid input option "
                f"for Sample {sample['Name']} e.g. RucioDID"
                )
        if 'RucioDID' in sample.keys():
            for did in sample['RucioDID'].split(","):
                if len(did.split(":")) != 2:
                    raise ValueError(
                        f"Sample {sample['Name']} "
                        f"- RucioDID {did} is missing the scope"
                        )
        # if ('Tree' in sample) and \
        #         ('uproot' not in config['General']['ServiceXName'].lower()):
        #     raise KeyError(
        #         f"Tree in Sample {sample['Name']} "
        #         "is only for uproot backend type"
        #         )
        if 'Columns' in sample and 'FuncADL' in sample:
            raise KeyError(
                f"Sample {sample['Name']} - Use one type of query per sample: "
                "Columns for TCut and FuncADL for func-adl"
                )
        if 'FuncADL' in sample and 'Filter' in sample:
            raise KeyError(
                f"Sample {sample['Name']} - "
                "You cannot use Filter with func-adl query"
                )

    log.debug("config looks okay")
    return True
